/* module for driving a LCD display    R. Bleher 1998,1999 <dk7in@qsl.net>
 * the idea and parts of this software were copied from other packages:
 *   lcd-0.12 by Nils Faerber and Michael Engel
 *
 * I'm using an 4x20 display, connected to a parallel port (8-bit)
 * See display.txt and hardware.h for more info
   23.10.99  rb  updated for 2.2.x kernel
*/
#include <linux/module.h>          /* MOD_INC_USE_COUNT MOD_DEC_USE_COUNT */

#if LINUX_VERSION_CODE >= 0x020100
#include <asm/uaccess.h>           /* for put_user and get_user */
#else
#include <asm/segment.h>           /* get_user (2.0.x) */
#include <linux/fs.h>
#endif

#include <linux/ioport.h>          /* release_region check_region request_region */
#include <asm/io.h>                /* outb */
#include <linux/delay.h>           /* udelay */

/* #include <asm/delay.h> */
/* #include <string.h> */          /* debug msgs ? */

#include "version.h"
#include "hardware.h"
#include "lcd.h"
#include "c_table.h"

/*
 * TODO: /proc entry showing the used port addresses
 */

static int  lcd_usage=0;
static char lcd_y=0,lcd_x=0;
static char lcd_i=0;               /* start with low intensity */

/***************************************************************************
* wr_ctrl()                                                                *
*                                                                          *
* Writes a char to the control register of the LCD display                 *
* with a correction mask for inverted bits                                 *
* Parameter: char data: data to write                                      *
* Return:    void                                                          *
***************************************************************************/
void wr_ctrl(char data)
{
  outb((data|lcd_i)^(CTRL_MASK),LCD_CTRL);
  udelay(MIN_DLY);
}

/***************************************************************************
* wr_data()                                                                *
*                                                                          *
* Writes a char to the data register of the LCD display                    *
* with a correction mask for inverted bits                                 *
* Parameter: char data: data to write                                      *
* Return:    void                                                          *
***************************************************************************/
void wr_data(char data)
{
  outb(data^(DATA_MASK),LCD_DATA);
  udelay(MIN_DLY);
}

/***************************************************************************
* write_command()                                                          *
*                                                                          *
* Set to instruction mode and write a command byte                         *
* wait until command is executed (busy time)                               *
* Parameter: char command: command to write                                *
* Return:    void                                                          *
***************************************************************************/
void write_command(char command)
{
  wr_ctrl(LCD_EN);                       /* RS=0 EN=1 */
  wr_data(command);                      /* write cmd */
  wr_ctrl(0x00);                         /* RS=0 EN=0 */
  wr_ctrl(LCD_EN);                       /* RS=0 EN=1 */
  if (command<=3)                        /* Clear Display and */
    udelay(MAX_DLY);                     /*   Cursor Home need more time */
  else
    udelay(DEF_DLY);                     /* busy time */
}

/***************************************************************************
* write_data()                                                             *
*                                                                          *
* Set to data mode and write a data byte                                   *
* wait until data is written to display (busy time)                        *
* for character translation see c_table.h                                  *
* Parameter: char data: data to write                                      *
* Return:    void                                                          *
***************************************************************************/
void write_data(char data)
{
  wr_ctrl(LCD_RS|LCD_EN);                /* RS=1 EN=1 */
  wr_data(c_table[(unsigned char)data]); /* write translated data */
  wr_ctrl(LCD_RS);                       /* RS=1 EN=0 */
  wr_ctrl(LCD_EN);                       /* RS=0 EN=1 */
  udelay(DEF_DLY);                       /* busy time */
}

/***************************************************************************
* set_light()                                                              *
*                                                                          *
* Set current value of LED backlight intensity                             *
* Parameter: void                                                          *
* Return:    void                                                          *
***************************************************************************/
void set_light(void)
{
  wr_ctrl(LCD_EN);                       /* RS=0 EN=1 */
}

/***************************************************************************
* init_lcd()                                                               *
*                                                                          *
* Resets LCD panel and sets initial mode, i.e.                             *
*   underline cursor, not blinking, one line mode, 5x7 small font          *
* Parameter: void                                                          *
* Return:    void                                                          *
***************************************************************************/
void init_lcd(void)
{
  udelay(15000);
  write_command(0x38);    /* Set 8-bit mode */
  udelay(4200);
  write_command(0x38);    /* Set 8-bit mode */
  udelay(4200);
  write_command(0x38);    /* Set 8-bit mode */
  udelay(4200);
  write_command(0x38);    /* Set 8-bit mode, font, matrix */
  write_command(0x08);    /* Set display off */
  write_command(0x0c);    /*15*/   /* Set display on, cursor off */
  write_command(0x06);    /* Set entry mode, cursor moves right, shift display */
  write_command(0x01);    /* Clear display */
}

/*  write_command(0x40); */   /* Set CG RAM address */
/*  write_command(0x80); */   /* Set DD RAM address */

/***************************************************************************
* set_cursor_pos()                                                             *
*                                                                          *
* Sets cursor to a certain position                                        *
* Parameter: unsigned char x:  column [0-19]                               *
*            unsigned char y:  line   [0- 3]                               *
* Return:    void                                                          *
***************************************************************************/
void set_cursor_pos(unsigned char x, unsigned char y)
{
  unsigned char val;

  if (x>(LCD_COLS-1)) {          /* some sanity checking */
    x=0; y++;
  }
  if (y>(LCD_ROWS-1)) y=0;

  val = x + (y%2) * 0x40;        /* calculate DD RAM address */
  if (y>=2) val += 20;
  
  write_command(0x80|val);       /* 0x80: addressing command */
  
  lcd_y = y;                     /* Update our private postion counter */
  lcd_x  = x;
}

/***************************************************************************
* Defines a custom character from 0-7...                                   *
*                                                                          *
* For input, values > 0 mean "on" and values <= 0 are "off".               *
*                                                                          *
* The input is just an array of characters...    bits                      *
***************************************************************************/
void define_char(int n, char *dat)
{
  int row, col;
  int letter;

  if(n < 0 || n > 7) return;
  if(!dat) return;

  write_command(0x40|n*8);       /* set CG RAM address */

  for(row=0; row<8; row++) {
    letter = 0;
    for(col=0; col<5; col++) {
      letter <<= 1;
      letter |= (dat[(row*5) + col] > 0);
    }
    write_data(letter);
  }
}


/***************************************************************************
* lcd_open()                                                               *
*                                                                          *
* Called when system call open is executed                                 *
* Parameter: see man open                                                  *
* Return:    see man open                                                  *
***************************************************************************/
static int lcd_open(struct inode *minode, struct file *file)
{
  if (lcd_usage!=0)
    return -EBUSY;

#ifdef MODULE
  MOD_INC_USE_COUNT;
#endif MODULE
  lcd_usage=1;
#ifdef DEBUG
  printk("LCD opened\n");
#endif

  return(0);
}

/***************************************************************************
* lcd_release()                                                            *
*                                                                          *
* Called when system call close is executed                                *
* Parameter: see man close                                                 *
* Return:    see man close                                                 *
***************************************************************************/
#if LINUX_VERSION_CODE >= 0x020100
static int lcd_release(struct inode *minode, struct file *mfile)
#else
static void lcd_release(struct inode *minode, struct file *mfile)
#endif
{
#ifdef MODULE
  MOD_DEC_USE_COUNT;
#endif MODULE
  lcd_usage=0;
#ifdef DEBUG
  printk("LCD module closed\n");
#endif
#if LINUX_VERSION_CODE >= 0x020100
    return 0;
#endif
}

/***************************************************************************
* lcd_write()                                                              *
*                                                                          *
* Called when system call write is executed                                *
* Parameter: see man write                                                 *
* Return:    see man write                                                 *
***************************************************************************/
#if LINUX_VERSION_CODE >= 0x020100
static ssize_t lcd_write(struct file *mfile, const char *gdata, size_t length,
                         loff_t *ppos)
#else
static int lcd_write(struct inode *minode, struct file *mfile, const char *gdata, int length)
#endif
{
int i;
const char *data;
char c;

  data=gdata;

  for (i=0; i<length; i++) {
#if LINUX_VERSION_CODE >= 0x020100
    get_user(c,(char *)(data+i));
#else
    c=get_user((char *)(data+i));
#endif LINUX_VERSION_CODE >= 0x020100
#ifdef DEBUG
    printk("LCD: starting %d\n",i);
    printk("LCD: which is '%c'\n",c);
#endif
    if (c==0x0a) {              /* line feed */
      lcd_y++;
      lcd_x=0;
      set_cursor_pos(lcd_x,lcd_y);
    } else
    if (c==0x0d) { /* carriage return */
      lcd_x=0;
      set_cursor_pos(lcd_x,lcd_y);
    } else
    if (c==0x1b) {              /* ESC */
      if ((i+1)<length) {
        i+=1;
#if LINUX_VERSION_CODE >= 0x020100
        get_user(c,(char *)(data+i));
#else
        c=get_user((char *)(data+i));
#endif LINUX_VERSION_CODE >= 0x020100
        if (c=='c') { 
          write_command(1);     /* clear display */
          lcd_x=0;
          lcd_y=0;
        } else
        if (c=='h') {           /* home cursor */
          write_command(0x80);
          lcd_x=0;
          lcd_y=0;
        } else
        if (c=='R') {
          init_lcd();           /* Reset */
          lcd_x=0;
          lcd_y=0;
        } else
        if (c=='H') {           /* High intensity */
          lcd_i=LCD_BL;
          set_light();
        } else
        if (c=='L') {           /* Low intensity */
          lcd_i=0;
          set_light();
        } else
        if (c=='U') {           /* User character */
           write_command(0x40|0*8);    // CG RAM # 0, small heart
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x0e);           //   - # # # -
           write_data(0x0e);           //   - # # # -
           write_data(0x0e);           //   - # # # -
           write_data(0x04);           //   - - # - -
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -

           write_command(0x40|1*8);    // CG RAM # 1, big heart
           write_data(0x00);           //   - - - - -
           write_data(0x0a);           //   - # - # -
           write_data(0x1f);           //   # # # # #
           write_data(0x1f);           //   # # # # #
           write_data(0x1f);           //   # # # # #
           write_data(0x0e);           //   - # # # -
           write_data(0x04);           //   - - # - -
           write_data(0x00);           //   - - - - -

           write_command(0x40|2*8);    // CG RAM # 2, Answering machine
           write_data(0x00);           //   - - - - -
           write_data(0x1f);           //   # # # # #
           write_data(0x0a);           //   - # - # -
           write_data(0x04);           //   - - # - -
           write_data(0x0e);           //   - # # # -
           write_data(0x0a);           //   - # - # -
           write_data(0x0e);           //   - # # # -
           write_data(0x00);           //   - - - - -

           write_command(0x40|3*8);    // CG RAM # 3, Fax
           write_data(0x00);           //   - - - - -
           write_data(0x10);           //   # - - - -
           write_data(0x1e);           //   # # # # -
           write_data(0x11);           //   # - - - #
           write_data(0x11);           //   # - - - #
           write_data(0x0f);           //   - # # # #
           write_data(0x01);           //   - - - - #
           write_data(0x00);           //   - - - - -

           write_command(0x40|4*8);    // CG RAM # 4, Mail
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x1f);           //   # # # # #
           write_data(0x15);           //   # - # - #
           write_data(0x15);           //   # - # - #
           write_data(0x1f);           //   # # # # #
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -

           write_command(0x40|5*8);    // CG RAM # 5, AdHoc Msg
           write_data(0x1c);           //   # # # - -
           write_data(0x14);           //   # - # - -
           write_data(0x1c);           //   # # # - -
           write_data(0x14);           //   # - # - -
           write_data(0x05);           //   - - # - #
           write_data(0x07);           //   - - # # #
           write_data(0x05);           //   - - # - #
           write_data(0x00);           //   - - - - -

           write_command(0x40|6*8);    // CG RAM # 6, 
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x0c);           //   - # # - -
           write_data(0x04);           //   - - # - -
           write_data(0x08);           //   - # - - -

           write_command(0x40|7*8);    // CG RAM # 7,
           write_data(0x1f);           //   - - - - -
           write_data(0x15);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x00);           //   - - - - -
           write_data(0x11);           //   - - - - -
           write_data(0x1b);           //   - - - - -
           write_data(0x1f);           //   - - - - -

           set_cursor_pos(lcd_x,lcd_y);

#ifdef rbxx
          if ((i+9)<length) {
            i+=1;
            n=get_user((char *)(data+i));
            if((n >= 0) && (n < 8)) {
              write_command(0x40|n*8);    /* set CG RAM address */
              for(row=0; row<8; row++) {
                i+=1;
                write_data(get_user((char *)(data+i)));
              }
              set_cursor_pos(lcd_x,lcd_y);
            } else
              i+=8;
          }
#endif
        }
      }
    } else {                    /* normal text */
        write_data(c);
        lcd_x++;
        if (lcd_x>(LCD_COLS-1)) {
          lcd_x=0;
          lcd_y++;
          set_cursor_pos(lcd_x,lcd_y);
        }
    }
  }
  return(length);
}

#ifdef MY_IOCTL
/***************************************************************************
* lcd_ioctl()                  from lcd-0.5.5                              *
*                                                                          *
* Called when system call iotcl is executed                                *
* Parameter: see man iotcl                                                 *
* Return:    see man iotcl                                                 *
***************************************************************************/
static int lcd_ioctl(struct inode *inode, struct file *file,
            unsigned int cmd, unsigned long arg)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int retval = 0;

#ifdef LCD_DEBUG
   printk(KERN_DEBUG "lcd%d ioctl, cmd: 0x%x, arg: 0x%lx\n", minor, cmd, arg);
#endif
   
     switch ( cmd ) {
           
//       case LCD_S_CURSOR:         /* Turn cursor on/off */
//         if (arg)
//	   LCD_F(minor) |= CURSOR_ON;
//	 else
//	   LCD_F(minor) &= ~CURSOR_ON;
//         break;
         
//       case LCD_S_BLINK:          /* Turn blinking cursor on/off */
//         if (arg)
//	   LCD_F(minor) |= BLINK_ON;
//	 else
//	   LCD_F(minor) &= ~BLINK_ON;
//         break;
         
       case LCD_S_CURX:             /* Set cursor x-position */
         if(arg < 0)                         /* from EOL */
	   lcd_x = LCD_COLS + arg;
	 else
	   lcd_x = arg - 1;                  /* set X position */
	 if(lcd_x >= LCD_COLS)
	   lcd_x = LCD_COLS - 1;             /* max EOL */
	 if(lcd_x < 0)
	   lcd_x = 0;                        /* not negative */
	 update = 0;
	 set_cursor_pos(lcd_x, lcd_y);
         break;
         
       case LCD_G_CURX: {         /* Get cursor x-position */
         int x = lcd_x + 1;
         put_user(x, (int *) arg);
         break;
       }
           
       case LCD_S_CURY:           /* Set cursor y-position */
         if(arg < 0)
	   lcd_y = LCD_ROWS + arg;
	 else
	   lcd_y = arg-1;
	 if(lcd_y >= LCD_ROWS)
	   lcd_y = LCD_ROWS - 1;
	 if(lcd_y < 0)
	   lcd_y = 0;
	 update = 0;
	 set_cursor_pos(lcd_x, lcd_y);
         break;
         
       case LCD_G_CURY: {         /* Get cursor y-position */
         int y = lcd_y + 1;
         put_user(y, (int *) arg);
         break;
       }
           
       case LCD_S_CHAR: {                       /* Set user chars */
         struct char_def defs_src;              /* bytes */
	 int nr, row;

         copy_from_user((char *) &defs_src, (char *) arg, sizeof(struct char_def));
         nr = defs_src.char_nr;
         if((nr >= 0) && (nr < 8)) {
           write_command(0x40|nr*8);            /* set CG RAM address */
           for(row=0; row<8; row++) {
             write_data(defs_src.data[row]);
           }
         }
         else {
           retval = -EINVAL;
	 }
         break;
       }

//       case LCDCLEAR: {           /* Clear display, move cursor to (1,1) */
//         char *ptr;
//         int j;
//
//         write_ctrl(minor, 0x01);
//
//         if(NULL != (ptr = LCD_BUF(minor)))
//         for(j=0; j<(LCD_ROWS(minor) * LCD_COLS(minor)); j++)
//         *ptr++ = 0x20;
//         lcd_x(minor) = 0;
//	 lcd_y(minor) = 0;
//         LCD_F(minor) |= BLINK_ON;
//         LCD_F(minor) |= CURSOR_ON;
//         LCD_F(minor) |= INCREMENT;
//         LCD_F(minor) &= ~RAW_ON; /* switch to cooked */
//         break;
//       }

//       case LCDRESET:             /* Reset display */
//         lcd_reset();
//         break;

       default:
         retval = -EINVAL;
     }
   return retval;
}
#endif MY_IOCTL

/***************************************************************************
* lcd_read()                                                               *
*                                                                          *
* Called when system call read is executed                                 *
* Parameter: see man read                                                  *
* Return: see man read                                                     *
***************************************************************************/
#if LINUX_VERSION_CODE >= 0x020100
static ssize_t lcd_read(struct file *mfile, char *data, size_t length,
                        loff_t *ppos)
#else
int lcd_read(struct inode *inode, struct file *mfile, char *data, int length)
#endif LINUX_VERSION_CODE >= 0x020100
{
#ifdef DEBUG
  printk("LCD: Trying to read readonly device\n");
#endif
  return(0);
/*
  sprintf(data,"%d %d\x004",lcd_x,lcd_y);
  return(strlen(data));
*/
}

/***************************************************************************
* cleanup_module()                                                         *
*                                                                          *
* Called when module is removed                                            *
* Parameter: void                                                          *
* Return:    void                                                          *
***************************************************************************/
#ifdef MODULE
void cleanup_module(void)
{
  release_region(LCD_DATA,3);
  /* printk("lcd: I/O region freed\n"); */
  if (unregister_chrdev(LCD_MAJOR,"lcd"))
    printk(KERN_INFO "lcd: module cleanup failed\n");
  else
    printk(KERN_INFO "lcd: driver unloaded\n");
}
#endif MODULE

#ifdef MODULE
#define lcd_init init_module
#endif MODULE

/***************************************************************************
* lcd_init()                                                               *
*                                                                          *
* Called when module is initialized after it has been loaded               *
* Parameter: void                                                          *
* Return:    int: 0 if no Error else < 0                                   *
***************************************************************************/
int lcd_init(void)
{
  static struct file_operations lcd_fops={
      NULL,                   /* lseek   */
      lcd_read,               /* read    */
      lcd_write,              /* write   */
      NULL,                   /* readdir */
      NULL,                   /* select  / poll */
      NULL,                   /* ioctl   */
      NULL,                   /* mmap    */
      lcd_open,               /* open    */
#if LINUX_VERSION_CODE >= 0x020100
      NULL,                   /* flush   */
#endif
      lcd_release,            /* release */
      NULL,                   /* fsync   */
      NULL,                   /* fasync  */
      NULL,                   /* check_media_change */
      NULL                    /* revalidate */
  };
  if (register_chrdev(LCD_MAJOR,"lcd",&lcd_fops)) {
      printk("lcd: unable to get major %d\n", LCD_MAJOR);
      return -EIO;
  }
  printk(KERN_INFO "LCD driver V%s, R.Bleher\n",LCD_MODULE_VERSION);
  if (!check_region(LCD_DATA,3)) {
    request_region(LCD_DATA,3,"lcd");
    printk(KERN_INFO "lcd: at address 0x%04x\n",LCD_DATA);
    lcd_usage=0;
    init_lcd();
    lcd_y=0;
    lcd_x=0;
  } else
  {
    release_region(LCD_DATA,3);
    unregister_chrdev(LCD_MAJOR,"lcd");
    printk(KERN_INFO "lcd: could not get 3 addresses from 0x%04x\n",LCD_DATA);
  }
  return 0;
}
